;; open X86

(* Generate and indirect address offset from EBP *)
let ebp_offset (amt:int32) : opnd =
  Ind{i_base = Some Ebp;
      i_iscl = None;
      i_disp = Some (DImm amt)} 

(* Generate a stack offset for a given argument index.
 * Arguments are numbered 1,2,...,n
 *)
let argument (index:int) : opnd =
  let amt = Int32.of_int ((index + 1) * 4) in
    ebp_offset amt

(* Generate a stack offset for a given local variable.
 * Locals are numbered 1,2,...,j
 *)
let local (index:int) : opnd =
  let amt = Int32.of_int (index * -4) in
    ebp_offset amt

let fn_entry : insn list = [
  Push(ebp);
  Mov(ebp, esp);
]

let fn_exit : insn list = [
  Pop(ebp);
  Ret;
]

(* Create a function body with n local variables.
 * Assumes that the body code places the return result in EAX
 *)
let fn_code (n:int) (body:insn list) : insn list =
  let temp_size = Int32.of_int (n*4) in
    fn_entry @                   (* save ebp and create frame ptr *)
    [Sub(esp, i32 temp_size)] @  (* allocate n temporary variables *)
    body @                          (* function body *)
    [Add(esp, i32 temp_size)] @  (* deallocate the n temporaries *)
    fn_exit
  
(* Create code that calls the function fn with arguments
 *   [a1;a2;a3]   
 *)
let call (fn:string) (args:opnd list) : insn list =
  let arg_size = Int32.of_int (4 * (List.length args)) in
  let pushes = List.fold_left (fun ps x -> Push(x)::ps) [] args in
    pushes @
    [Call(Lbl (mk_lbl_named fn));
     Add(esp, i32 arg_size)]


(******************************************)
(*   Redo the example from handcoded1.ml  *)
(******************************************)

let square_block = 
  mk_block "square"
    (fn_code 0 [Mov(eax, argument 1);
		Imul(Eax, eax);])

let f_block =
  mk_block "f"
    (fn_code 0 
       ([Mov(eax, argument 2);
	 Mov(ebx, argument 1);
	 Sub(eax, ebx);]
	@ (call "square" [eax])))

let main_block = {
  global = true;  (* Make this block visible to the linker *)
  label = mk_lbl_named "_program";
  insns = fn_code 0 (call "f" [i32 10l; i32 5l])
}

let program =
  List.map (fun x -> Cunit.Code x) [main_block; f_block; square_block] 

let _ = Cunit.output_cunit program stdout

