;; open X86


(* Generate and indirect address offset from EBP *)
let ebp_offset (amt:int32) : opnd =
  Ind{i_base = Some Ebp;
      i_iscl = None;
      i_disp = Some (DImm amt)} 

(* Generate a stack offset for a given argument index.
 * Arguments are numbered 1,2,...,n
 *)
let argument (index:int) : opnd =
  let amt = Int32.of_int ((index + 1) * 4) in
    ebp_offset amt

(* Generate a stack offset for a given local variable.
 * Locals are numbered 1,2,...,j
 *)
let local (index:int) : opnd =
  let amt = Int32.of_int (index * -4) in
    ebp_offset amt

let fn_entry : insn list = [
  Push(ebp);
  Mov(ebp, esp);
]

let fn_exit : insn list = [
  Pop(ebp);
  Ret;
]

(* Create a function body with n local variables.
 * Assumes that the body code places the return result in EAX
 *)
let fn_code (n:int) (body:insn list) : insn list =
  let temp_size = Int32.of_int (n*4) in
    fn_entry @                   (* save ebp and create frame ptr *)
    [Sub(esp, i32 temp_size)] @  (* allocate n temporary variables *)
    body @                          (* function body *)
    [Add(esp, i32 temp_size)] @  (* deallocate the n temporaries *)
    fn_exit
  
(* Create code that calls the function fn with arguments
 *   [a1;a2;a3]   
 *)
let call (fn:string) (args:opnd list) : insn list =
  let arg_size = Int32.of_int (4 * (List.length args)) in
  let pushes = List.fold_left (fun ps x -> Push(x)::ps) [] args in
    pushes @
    [Call(Lbl (mk_lbl_named fn));
     Add(esp, i32 arg_size)]


(*********************)
(* Factorial example *)
(*********************)

let fact_entry_block = 
  mk_block "fact"
    (fn_code 1 [Mov(eax, argument 1);
		Cmp(eax, i32 1l);
		J(Sgt, mk_lbl_named "else_branch");
		Mov(eax, i32 1l);])

let else_branch_block =
  mk_block "else_branch"
    ([Mov(local 1, eax);
      Sub(eax, i32 1l)]
     @ (call "fact" [eax]) @
      [Mov(ebx, local 1);
       Imul(Eax, ebx);
       Add(esp, i32 4l)]
     @ fn_exit)
       
let program = {
  global = true;
  label  = mk_lbl_named "_program";
  insns = fn_code 0 (call "fact" [i32 5l])
}

let program = 
  List.map (fun x -> Cunit.Code x) (fact_entry_block::else_branch_block::program::[])

let _ = Cunit.output_cunit program stdout


    
